#include "stdafx.h"
#include "SCardWrappers.h"
#include <array>
#include <wincrypt.h>



#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#ifndef SCARD_PROVIDER_CARD_MODULE
#define SCARD_PROVIDER_CARD_MODULE 0x80000001
#endif

#ifndef SCARD_AUDIT_CHV_FAILURE
#define SCARD_AUDIT_CHV_FAILURE 0x0
#endif

enum class AttributeType
{
  BLOB_OR_INTEGER_ATTRIBUTE = 0,
  ASCII_STRING_ATTRIBUTE = 1,
  STRING_ATTRIBUTE = 2,
  BLOB_ATTRIBUTE = 3,
};


SCard::String BinaryToString(_In_ const SCard::ByteArray& input)
{
  DWORD dwString{0};
  const BYTE* pData = input.data();
#pragma warning(suppress: 26472)
  const auto dwSize = static_cast<DWORD>(input.size());
  if (!CryptBinaryToString(pData, dwSize, CRYPT_STRING_HEXRAW | CRYPT_STRING_NOCRLF, nullptr, &dwString))
    return {};
  std::vector<TCHAR> sTemp{dwString, std::allocator<TCHAR>{}};
  if (!CryptBinaryToString(pData, dwSize, CRYPT_STRING_HEXRAW | CRYPT_STRING_NOCRLF, sTemp.data(), &dwString))
    return {};
  return sTemp.data();
}

SCard::String AttributeBinaryToString(_In_ const SCard::ByteArray& input)
{
  //What will be the return value from this method
  SCard::String sReturn;

  const auto nInputLength{input.size()};
  if (nInputLength == 1)
  {
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
#pragma warning(suppress: 26472 26446)
    ss << static_cast<int>(input[0]);
    sReturn = ss.str();
  }
  else if (nInputLength == 2)
  {
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
#pragma warning(suppress: 26490 26429)
    const auto* pwValue{reinterpret_cast<const WORD*>(input.data())};
    ss << *pwValue;
    sReturn = ss.str();
  }
  else if (nInputLength == 4)
  {
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
#pragma warning(suppress: 26490 26429)
    const auto* pdwValue{reinterpret_cast<const DWORD*>(input.data())};
    ss << *pdwValue;
    sReturn = ss.str();
  }
  else
    sReturn = BinaryToString(input);
  return sReturn;
}

SCard::String AttributeBinaryStringToString(_In_ const SCard::ByteArray& input)
{
  //What will be the return value from this method
  SCard::String sReturn;

  const auto nInputLength{input.size()};
  if (nInputLength)
  {
    SCard::ByteArray inputCopy{input};
    inputCopy.push_back(0x00); //Don't trust that the data will be null terminated
  #ifdef _UNICODE
    inputCopy.push_back(0x00); //Don't trust that the data will be null terminated
  #endif //#ifdef _UNICODE
#pragma warning(suppress: 26490)
    auto pszValue{reinterpret_cast<LPCTSTR>(inputCopy.data())};
    sReturn = pszValue;
  }
  return sReturn;
}

SCard::String AttributeASCIIBinaryStringToString(_In_ const SCard::ByteArray& input)
{
  //What will be the return value from this method
  SCard::String sReturn;

  const auto nInputLength{input.size()};
  if (nInputLength)
  {
    SCard::ByteArray inputCopy{input};
    inputCopy.push_back(0x00); //Don't trust that the data will be null terminated
#pragma warning(suppress: 26490)
    auto pszValue{reinterpret_cast<LPCSTR>(inputCopy.data())};
    sReturn = ATL::CA2T{pszValue};
  }
  return sReturn;
}

_Success_(return != false) bool StringToBinary(_In_ const SCard::String& sInput, _Out_ SCard::ByteArray& output)
{
  DWORD dwBinary{0};
  const TCHAR* psInput = sInput.c_str();
  if (!CryptStringToBinary(psInput, 0, CRYPT_STRING_HEXRAW, nullptr, &dwBinary, nullptr, nullptr))
    return false;
  output.resize(dwBinary);
  if (!CryptStringToBinary(psInput, 0, CRYPT_STRING_HEXRAW, output.data(), &dwBinary, nullptr, nullptr))
    return false;
#pragma warning(suppress: 26489)
  output.resize(dwBinary);
  return true;
}

void PrintByteArray(_In_ const SCard::ByteArray& input, _In_ AttributeType type)
{
  SCard::String sValue;
  switch (type)
  {
    case AttributeType::BLOB_ATTRIBUTE:
    {
      sValue = BinaryToString(input);
      break;
    }
    case AttributeType::BLOB_OR_INTEGER_ATTRIBUTE:
    {
      sValue = AttributeBinaryToString(input);
      break;
    }
    case AttributeType::ASCII_STRING_ATTRIBUTE:
    {
      sValue = AttributeASCIIBinaryStringToString(input);
      break;
    }
    case AttributeType::STRING_ATTRIBUTE:
    {
      sValue = AttributeBinaryStringToString(input);
      break;
    }
    default:
    {
      ASSERT(false);
      break;
    }
  }
  _tprintf(_T("%s"), sValue.c_str());
}

void PrintATR(_Inout_ SCard::ByteArray& ATR)
{
  _tprintf(_T("ATR details: "));
  PrintByteArray(ATR, AttributeType::BLOB_ATTRIBUTE);
  _tprintf(_T("\n"));

  SCard::CATR parser;
  SCard::CATR::ParseError error{SCard::CATR::ParseError::NoError};
  const bool bParse{parser.Parse(ATR, error)};
  if (bParse)
  {
    //TS details
#pragma warning(suppress: 26472)
    _tprintf(_T("TS initial character value: 0x%02X\n"), static_cast<int>(parser.m_byTS));
    if (parser.m_bDirectConvention)
      _tprintf(_T(" TS specifies the direct convention\n"));
    else
      _tprintf(_T(" TS specifies the inverse convention\n"));

    //T0 details
#pragma warning(suppress: 26472)
    _tprintf(_T("T0 Format character value: 0x%02X\n"), static_cast<int>(parser.m_byT0));
    _tprintf(_T(" Y1: 0x%1X\n"), parser.m_nY1);
    _tprintf(_T(" K (Number of Historical characters): %d\n"), parser.m_nK);

    //TA1 to TD4 details
    if (parser.m_bTA1Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TA1 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTA1));
      _tprintf(_T(" FI (Clock rate conversion factor): 0x%1X\n"), parser.m_nFI);
      _tprintf(_T(" DI (Bit rate adjustment factor): 0x%1X\n"), parser.m_nDI);
      if (parser.m_nF == -1)
      {
        _tprintf(_T(" F (clock rate conversion factor) is RFU, defaulting to 372\n"));
        _tprintf(_T(" F (clock rate conversion factor) is RFU, defaulting fMax to 4 Mhz\n"));
        parser.m_fMax = 4;
        parser.m_nF = 372;
      }
      _tprintf(_T(" F (Clock rate conversion factor): %d\n"), parser.m_nF);
      if (parser.m_nD == -1)
      {
        _tprintf(_T(" D (Bit rate adjustment factor) is RFU, defaulting to 1\n"));
        parser.m_nD = 1;
      }
      _tprintf(_T(" D (Bit rate adjustment factor): %d\n"), parser.m_nD);
      _tprintf(_T(" fMax: %f Mhz\n"), parser.m_fMax);
      _tprintf(_T(" %d clock cycles per ETU\n"), parser.m_nF/parser.m_nD);
    }
    if (parser.m_bTB1Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TB1 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTB1));
      _tprintf(_T(" PI1 (EEPROM Programming parameter): %d\n"), parser.m_nPI1);
      _tprintf(_T(" II (EEPROM Programming parameter): %d\n"), parser.m_nII);
      if (parser.m_byTB1 == 0)
        _tprintf(_T(" Vpp is not electrically connected\n"));
    }
    if (parser.m_bTC1Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TC1 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTC1));
#pragma warning(suppress: 26472)
      _tprintf(_T(" N (Extra guard time): %d\n"), static_cast<int>(parser.m_byTC1));
    }
    if (parser.m_bTD1Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TD1 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTD1));
      _tprintf(_T(" Y2: 0x%1X\n"), parser.m_nY2);
      _tprintf(_T(" T (Transmission Protocol): %d\n"), parser.m_nProtocol1);
    }
    if (parser.m_bTA2Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TA2 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTA2));
      if (parser.m_bSwitchBetweenNegotiableModeAndSpecificModePossible)
        _tprintf(_T(" Switching between negotiable mode and specific mode is possible\n"));
      else
        _tprintf(_T(" Switching between negotiable mode and specific mode is not possible\n"));
      if (parser.m_bTransmissionCharactersImplicitlyDefined)
        _tprintf(_T(" Transmission parameters implicitly defined in the interface characters\n"));
      else
        _tprintf(_T(" Transmission parameters explicitly defined in the interface characters\n"));
      _tprintf(_T(" Protocol: %d\n"), parser.m_nTA2Protocol);
    }
    if (parser.m_bTB2Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TB2 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTB2));
      _tprintf(_T(" PI2/IFSC (EEPROM Programming Voltage): %f volts\n"), parser.m_nPI2 * 0.1);
    }
    if (parser.m_bTC2Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TC2 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTC2));
#pragma warning(suppress: 26472)
      _tprintf(_T(" WI (Work waiting time): %d\n"), static_cast<int>(parser.m_byTC2));
    }
    if (parser.m_bTD2Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TD2 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTD2));
      _tprintf(_T(" Y3: 0x%1X\n"), parser.m_nY3);
      _tprintf(_T(" T (Transmission Protocol): %d\n"), parser.m_nProtocol2);
    }
    if (parser.m_bTA3Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TA3 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTA3));
    }
    if (parser.m_bTB3Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TB3 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTB3));
      _tprintf(_T(" CWI: %d\n"), parser.m_nTB3CWI);
      _tprintf(_T(" BWI: %d\n"), parser.m_nTB3BWI);
    }
    if (parser.m_bTC3Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TC3 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTC3));
      if (parser.m_bTC3CRCUsed)
        _tprintf(_T(" CRC Error detection code used\n"));
      else
        _tprintf(_T(" LRC Error detection code used\n"));
    }
    if (parser.m_bTD3Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TD3 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTD3));
      _tprintf(_T(" Y4: 0x%1X\n"), parser.m_nY4);
      _tprintf(_T(" T (Transmission Protocol): %d\n"), parser.m_nProtocol3);
    }
    if (parser.m_bTA4Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TA4 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTA4));
    }
    if (parser.m_bTB4Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TB4 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTB4));
      _tprintf(_T(" CWI: %d\n"), parser.m_nTB4CWI);
      _tprintf(_T(" BWI: %d\n"), parser.m_nTB4BWI);
    }
    if (parser.m_bTC4Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TC4 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTC4));
      if (parser.m_bTC4CRCUsed)
        _tprintf(_T(" CRC Error detection code used\n"));
      else
        _tprintf(_T(" LRC Error detection code used\n"));
    }
    if (parser.m_bTD4Present)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TD4 interface character value: 0x%02X\n"), static_cast<int>(parser.m_byTD4));
      _tprintf(_T(" Y5: 0x%1X\n"), parser.m_nY5);
      _tprintf(_T(" T (Transmission Protocol): %d\n"), parser.m_nProtocol4);
    }

    //Historical characters details
    _tprintf(_T("Historical characters: "));
#pragma warning(suppress: 26472 26489)
    const auto nHistoricalCharacters{static_cast<int>(parser.m_HistoricalCharacters.size())};
    for (int i=0; i<nHistoricalCharacters; i++)
#pragma warning(suppress: 26446 26472 26489)
      _tprintf(_T("%02X"), static_cast<int>(parser.m_HistoricalCharacters[i]));
    _tprintf(_T("\n"));

    for (int i=0; i<nHistoricalCharacters; i++)
#pragma warning(suppress: 26446 26472 26489)
      _tprintf(_T(" T%d: %02X\n"), i+1, static_cast<int>(parser.m_HistoricalCharacters[i]));

    //Historical characters as ASCII details
#pragma warning(suppress: 26489)
    SCard::ByteArray historicalCharactersAsASCII(parser.m_HistoricalCharacters);
    historicalCharactersAsASCII.push_back(0);
#pragma warning(suppress: 26490)
    printf(" as ASCII: \"%s\"\n", reinterpret_cast<char*>(historicalCharactersAsASCII.data()));

    //Category indicator in T1
    if (nHistoricalCharacters)
    {
#pragma warning(suppress: 26446)
      switch (parser.m_HistoricalCharacters[0])
      {
        case 0x00:
        {
          _tprintf(_T(" Category indicator indicates status information is located at the end of the historical characters\n"));
          break;
        }
        case 0x10:
        {
          _tprintf(_T(" Category indicator indicates a reference to a DIR file (directory file)\n"));
          break;
        }
        case 0x80:
        {
          _tprintf(_T(" Category indicator indicates status information is contained in a compact TLV-coded data object\n"));
          break;
        }
        default:
        {
          _tprintf(_T(" Unknown Category indicator\n"));
          break;
        }
      }
    }

    //TCK details
    if (parser.m_bTCKPresent)
    {
#pragma warning(suppress: 26472)
      _tprintf(_T("TCK Checksum: %02X\n"), static_cast<int>(parser.m_byTCK));
      if (parser.m_bTCKValid)
        _tprintf(_T(" Checksum is valid\n"));
      else
        _tprintf(_T(" Checksum is not valid\n"));
    }
  }
  else
  {
    switch (error)
    {
      case SCard::CATR::ParseError::TSNotFound:
      {
        _tprintf(_T("Failed to parse ATR, TS not found\n"));
        break;
      }
      case SCard::CATR::ParseError::UnknownTSValue:
      {
        _tprintf(_T("Failed to parse ATR, Unknown TS value\n"));
        break;
      }
      case SCard::CATR::ParseError::T0NotFound:
      {
        _tprintf(_T("Failed to parse ATR, T0 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TA1NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TA1 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TA2NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TA2 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TA3NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TA3 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TA4NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TA4 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TB1NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TB1 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TB2NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TB2 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TB3NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TB3 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TB4NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TB4 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TC1NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TC1 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TC2NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TC2 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TC3NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TC3 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TC4NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TC4 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TD1NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TD1 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TD2NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TD2 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TD3NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TD3 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TD4NotFound:
      {
        _tprintf(_T("Failed to parse ATR, TD4 not found\n"));
        break;
      }
      case SCard::CATR::ParseError::HistoricalCharactersNotFound:
      {
        _tprintf(_T("Failed to parse ATR, Historical Characters not found\n"));
        break;
      }
      case SCard::CATR::ParseError::TCKIsNotLast:
      {
        _tprintf(_T("Failed to parse ATR, TCK is not last character\n"));
        break;
      }
      default:
      {
        _tprintf(_T("Failed to parse ATR, Unknown Error\n"));
        break;
      }
    }
  }
}


int _tmain()
{
  //Create the context
  SCard::CContext ctx;
  LONG nRet{ctx.Create()};
  if (nRet != SCARD_S_SUCCESS)
  {
#pragma warning(suppress: 26472)
    _tprintf(_T("Failed to create SCard context, giving up!, Error:0x%08X\n"), static_cast<unsigned int>(nRet));
    return 1;
  }
  SCARDCONTEXT h{ctx.Detach()};
  SCard::CContext ctxm;
  ctxm.Attach(h);
  SCard::CContext ctx3{std::move(ctxm)};
  SCard::CContext ctx11;
  ctx11.Attach(ctx3.Detach());
#pragma warning(suppress: 26477 26486)
  ATLASSERT(ctx11.IsValid() == SCARD_S_SUCCESS);
  //SCard::CContext ctx5{ctx};
  //SCard::CContext ctx6{ctx5.Detach()};
  ctx.Attach(ctx11.Detach());

  SCard::CContext ctx4;
  SCard::StringArray readers;
  nRet = ctx4.ListReaders(SCARD_ALL_READERS, readers); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  SCard::String sFirstReader;
  SCard::String sValue;
  SCard::StringArray readers2;
  int i{0};
  for (const auto& reader : readers)
  {
#pragma warning(suppress: 26489)
    _tprintf(_T("Found Reader: %s\n"), reader.c_str());
    if (i == 0)
#pragma warning(suppress: 26486 26489)
      sFirstReader = reader;
#pragma warning(suppress: 26486 26489)
    nRet = ctx.GetReaderDeviceInstanceId(reader.c_str(), sValue);
    if (nRet == SCARD_S_SUCCESS)
#pragma warning(suppress: 26486)
      nRet = ctx.ListReadersWithDeviceInstanceId(sValue.c_str(), readers2); //NOLINT(clang-analyzer-deadcode.DeadStores)
    ++i;
  }

  SCard::StringArray readerGroups;
  nRet = ctx4.ListReaderGroups(readerGroups); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  for (const auto& readerGroup : readerGroups)
    _tprintf(_T("Found Reader group: %s\n"), readerGroup.c_str());

  //uncomment and modify the following line if you would like to connect to a specific smart card reader
  //sFirstReader = _T("SCM Microsystems Inc. SDI011G Smart Card Reader 0");

  if (sFirstReader.length() == 0)
  {
    _tprintf(_T("No readers found on this computer, giving up!\n"));
    return 1;
  }

  SCARD_READERSTATE srs{};
  srs.szReader = sFirstReader.c_str();
  srs.dwCurrentState = SCARD_STATE_UNAWARE;
#pragma warning(suppress: 26486)
  nRet = ctx.LocateCards(_T("Identity Device (NIST SP 800-73 [PIV])\0"), &srs, 1); //NOLINT(clang-analyzer-deadcode.DeadStores)

  SCARD_ATRMASK atrMask;
  atrMask.cbAtr = 23;
#pragma warning(suppress: 26485)
  memcpy_s(atrMask.rgbAtr, sizeof(atrMask.rgbAtr), "\x3B\xEE\x00\x00\x81\x31\xFE\x35\x00\x31\xC0\x65\xBC\xD0\x02\x01\x06\x71\xD6\x8C\x61\x22\x00", 23);
#pragma warning(suppress: 26485)
  memset(atrMask.rgbMask, 0xFF, sizeof(atrMask.rgbMask));
#pragma warning(suppress: 26486)
  nRet = ctx.LocateCardsByATR(&atrMask, 1, &srs, 1); //NOLINT(clang-analyzer-deadcode.DeadStores)

  const GUID MyGuid{0xABCDEF00, 0xABCD, 0xABCD, {0xAA, 0xBB, 0xCC, 0xDD, 0xAA, 0xBB, 0xCC, 0xDD}};
  std::array<BYTE, 5> MyATR{0xaa, 0xbb, 0xcc, 0x00, 0xdd};
  std::array<BYTE, 5> MyATRMask{0xff, 0xff, 0xff, 0x00, 0xff};
#pragma warning(suppress: 26472 26486)
  nRet = ctx.IntroduceCardType(_T("PJ's Card"), &MyGuid, nullptr, 0, MyATR.data(), MyATRMask.data(), static_cast<DWORD>(MyATR.size())); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

#pragma warning(suppress: 26486)
  nRet = ctx.ForgetCardType(_T("PJ's Card")); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

#pragma warning(suppress: 26486)
  nRet = ctx.IntroduceReader(_T("PJs Reader"), _T("Broadcom Corp Contacted SmartCard 0")); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

#pragma warning(suppress: 26486)
  nRet = ctx.ForgetReader(_T("PJs Reader")); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  SCard::StringArray cards;
  nRet = ctx4.ListCards(nullptr, nullptr, 0, cards); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  for (const auto& card: cards)
  {
    LPCTSTR pszCard{card.c_str()};
    _tprintf(_T("Found Card: %s\n"), pszCard);
#pragma warning(suppress: 26486)
    nRet = ctx.GetProviderName(pszCard, SCARD_PROVIDER_PRIMARY, sValue);
    if (nRet == SCARD_S_SUCCESS)
      _tprintf(_T(" Primary provider: %s\n"), sValue.c_str());
#pragma warning(suppress: 26486)
    nRet = ctx.GetProviderName(pszCard, SCARD_PROVIDER_CSP, sValue); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(nRet == SCARD_S_SUCCESS);
    _tprintf(_T(" CSP provider: %s\n"), sValue.c_str());
#pragma warning(suppress: 26486)
    nRet = ctx.GetProviderName(pszCard, SCARD_PROVIDER_KSP, sValue);
    if (nRet == SCARD_S_SUCCESS)
      _tprintf(_T(" KSP provider: %s\n"), sValue.c_str());
    nRet = ctx.GetProviderName(pszCard, SCARD_PROVIDER_CARD_MODULE, sValue); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(nRet == SCARD_S_SUCCESS);
    _tprintf(_T(" Card module: %s\n"), sValue.c_str());

    GUID gProviderId;
    nRet = ctx.GetProviderId(pszCard, &gProviderId); //NOLINT(clang-analyzer-deadcode.DeadStores)

    SCard::GUIDArray interfaces;
    nRet = ctx.ListInterfaces(pszCard, interfaces); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(nRet == SCARD_S_SUCCESS);
  }

  //Test out ATR parsing
  SCard::ByteArray ATR;
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146155620312e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146155620812e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1314615562081803012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1314615562081003012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1314615562081103012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1F146155620812e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1E146155620812e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146155620912e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3b"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb5"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb511"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F131"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1314615"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F131461556"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146155620"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F1314615562031"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146155620312e"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F55620312e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F6"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F62031"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F20312"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F131"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F6"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F62031"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e30"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F62031"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e30"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e3012"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e301234"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312e30123456"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF0"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF012"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF01234"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF0123456"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100813146155620312e301E"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100813146155620312e301F"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3C"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3B"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb511"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3BFA1300FF813180450031C173C00100009000B1"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3BFA1300"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3B0EBCE1CAAFC2DFBCADBCD3C3DCBFA8"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100813146155620312e301E0012"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bF51100F13146F55620312e3012345678"), ATR));
  PrintATR(ATR);
  VERIFY(StringToBinary(_T("3bb51100F13146F5F620312eF0123456781234567890"), ATR));
  PrintATR(ATR);

  _tprintf(_T("Checking state of the first reader for the next 10 seconds\n"));
  for (int j=0; j<40; j++)
  {
    srs.dwCurrentState = SCARD_STATE_UNAWARE;
    nRet = ctx.GetStatusChange(INFINITE, &srs, 1); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(nRet == SCARD_S_SUCCESS);

    if (srs.dwEventState & SCARD_STATE_IGNORE)
      _tprintf(_T("Reader ignored,"));
    if (srs.dwEventState & SCARD_STATE_CHANGED)
      _tprintf(_T("Reader state has changed,"));
    if (srs.dwEventState & SCARD_STATE_UNKNOWN)
      _tprintf(_T("Reader name is not recognized,"));
    if (srs.dwEventState & SCARD_STATE_UNAVAILABLE)
      _tprintf(_T("Reader state not available,"));
    if (srs.dwEventState & SCARD_STATE_EMPTY)
      _tprintf(_T("No card in the reader,"));
    if (srs.dwEventState & SCARD_STATE_PRESENT)
      _tprintf(_T("Card in the reader,"));
    if (srs.dwEventState & SCARD_STATE_ATRMATCH)
      _tprintf(_T("Matching ATR found,"));
    if (srs.dwEventState & SCARD_STATE_EXCLUSIVE)
      _tprintf(_T("Card in the reader is allocated for exclusive use by another application,"));
    if (srs.dwEventState & SCARD_STATE_INUSE)
      _tprintf(_T("Card in use in the reader but may be connected to in shared mode,"));
    if (srs.dwEventState & SCARD_STATE_MUTE)
      _tprintf(_T("There is an unresponsive card in the reader,"));

    if (srs.cbAtr)
    {
      _tprintf(_T("ATR of card: "));
      for (DWORD k=0; k<srs.cbAtr; k++)
#pragma warning(suppress: 26446 26472 26482)
        _tprintf(_T("%02X"), static_cast<int>(srs.rgbAtr[k]));
      _tprintf(_T("\n"));
    }
    else
      _tprintf(_T("\n"));

    //The current state is now the event state
    srs.dwCurrentState = srs.dwEventState;

    Sleep(250);
  }

  nRet = ctx.IntroduceReaderGroup(_T("PJs Group")); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  nRet = ctx.AddReaderToGroup(_T("Broadcom Corp Contacted SmartCard 0\0"), _T("PJs Group")); //NOLINT(clang-analyzer-deadcode.DeadStores)

  nRet = ctx.RemoveReaderFromGroup(_T("Broadcom Corp Contacted SmartCard 0\0"), _T("PJs Group")); //NOLINT(clang-analyzer-deadcode.DeadStores)

  nRet = ctx.ListReaderGroups(readerGroups); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  for (const auto& readerGroup : readerGroups)
    _tprintf(_T("Found Reader group: %s\n"), readerGroup.c_str());

  nRet = ctx.ForgetReaderGroup(_T("PJs Group")); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  nRet = ctx.Cancel(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  nRet = ctx.ForgetCardType(_T("DOES NOT EXIST")); //NOLINT(clang-analyzer-deadcode.DeadStores)

  nRet = ctx.Audit(SCARD_AUDIT_CHV_FAILURE); //NOLINT(clang-analyzer-deadcode.DeadStores)

  DWORD dwDeviceTypeId{0};
  SCard::String sDeviceInstanceId;
  SCard::ByteArray icon;
  nRet = ctx.GetDeviceTypeId(sFirstReader.c_str(), &dwDeviceTypeId); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nRet = ctx.GetReaderDeviceInstanceId(sFirstReader.c_str(), sDeviceInstanceId);
  if (nRet == SCARD_S_SUCCESS)
    nRet = ctx.ListReadersWithDeviceInstanceId(sDeviceInstanceId.c_str(), readers); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nRet = ctx.GetReaderIcon(sFirstReader.c_str(), icon); //NOLINT(clang-analyzer-deadcode.DeadStores)

  UUID uuid{};
  RPC_STATUS status = UuidCreate(&uuid); //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-but-set-variable)
#pragma warning(suppress: 26477)
  ATLASSUME(status == RPC_S_OK);
#pragma warning(suppress: 26465 26490 26492)
  nRet = ctx.WriteCache(&uuid, 1, const_cast<LPTSTR>(_T("PJ_Value1")), const_cast<LPBYTE>(reinterpret_cast<LPCBYTE>("1234")), 4);
  if (nRet == SCARD_S_SUCCESS)
  {
    std::array<BYTE, 4> byBuffer;
#pragma warning(suppress: 26472)
    auto dwSize{static_cast<DWORD>(byBuffer.size())};
#pragma warning(suppress: 26465 26492)
    nRet = ctx.ReadCache(&uuid, 1, const_cast<LPTSTR>(_T("PJ_Value1")), byBuffer.data(), &dwSize); //NOLINT(clang-analyzer-deadcode.DeadStores)
  }

  SCard::CCard card;
  _tprintf(_T("Attempting to connect to reader %s\n"), sFirstReader.c_str());
  nRet = card.Connect(ctx, sFirstReader.c_str());
  if (nRet != SCARD_S_SUCCESS)
  {
#pragma warning(suppress: 26472)
    _tprintf(_T("Could not connect to the card in the reader, giving up!, Error:0x%08X\n"), static_cast<unsigned int>(nRet));
    return 1;
  }

  SCARDHANDLE hCard{card.Detach()};
  card.Attach(hCard);
  SCard::CCard card2;
  card2 = std::move(card);
  card = std::move(card2);

  SCard::ByteArray Atr;
  DWORD dwState{0};
  DWORD dwProtocol{0};
  SCard::StringArray readerNames;
  nRet = card.Status(readerNames, &dwState, &dwProtocol, Atr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  switch (dwState)
  {
    case SCARD_ABSENT:
    {
      _tprintf(_T("Card absent\n"));
      break;
    }
    case SCARD_PRESENT:
    {
      _tprintf(_T("Card present\n"));
      break;
    }
    case SCARD_SWALLOWED:
    {
      _tprintf(_T("Card swallowed\n"));
      break;
    }
    case SCARD_POWERED:
    {
      _tprintf(_T("Power has power\n"));
      break;
    }
    case SCARD_NEGOTIABLE:
    {
      _tprintf(_T("Card reset and waiting PTS negotiation\n"));
      break;
    }
    case SCARD_SPECIFIC:
    {
      _tprintf(_T("Card has specific communication protocols set\n"));
      break;
    }
    default:
    {
      _tprintf(_T("Unknown or unexpected card state\n"));
      break;
    }
  }

  PrintATR(Atr);

  DWORD dwActiveProtocol{card.GetActiveProtocol()};
  UNREFERENCED_PARAMETER(dwActiveProtocol);

  LPBYTE pbAttr{nullptr};
  DWORD dwByte{SCARD_AUTOALLOCATE};
#pragma warning(suppress: 26490)
  card.GetAttrib(SCARD_ATTR_DEVICE_SYSTEM_NAME, reinterpret_cast<LPBYTE>(&pbAttr), &dwByte);
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  ctx.FreeMemory(pbAttr);

  nRet = card.BeginTransaction(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  nRet = card.EndTransaction(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  DWORD dwTransmitCount{0};
  nRet = card.GetTransmitCount(&dwTransmitCount); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Transmit Count: %u\n"), dwTransmitCount); //NOLINT(clang-diagnostic-format)

  nRet = card.Reconnect(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);

  SCard::ByteArray attrib;
  nRet = card.GetAttrib(SCARD_ATTR_ATR_STRING, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("ATR of card: "));
  for (const auto& ATRChar : attrib)
#pragma warning(suppress: 26472)
    _tprintf(_T("%02X"), static_cast<int>(ATRChar));
  _tprintf(_T("\n"));

  DWORD dwAttrib{0};
  nRet = card.GetAttrib(SCARD_ATTR_CHANNEL_ID, dwAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  if (dwAttrib & 0x80000)
    _tprintf(_T("Reader is SCSI based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else if (dwAttrib & 0x40000)
    _tprintf(_T("Reader is PS/2 Keyboard based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else if (dwAttrib & 0x20000)
    _tprintf(_T("Reader is Parallel port based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else if (dwAttrib & 0x10000)
    _tprintf(_T("Reader is Serial port based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else if (dwAttrib & 0x10000)
    _tprintf(_T("Reader is IDE based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else if (dwAttrib & 0x200000)
    _tprintf(_T("Reader is USB based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
  else
    _tprintf(_T("Reader Characteristics, Attribute value: 0x%08X\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

  nRet = card.GetAttrib(SCARD_ATTR_CHARACTERISTICS, dwAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  if (dwAttrib == 0)
    _tprintf(_T("Reader has no special characteristics\n"));
  else if (dwAttrib & 0x1)
    _tprintf(_T("Reader supports card swallowing\n"));
  else if (dwAttrib & 0x2)
    _tprintf(_T("Reader supports card ejection\n"));
  else if (dwAttrib & 0x4)
    _tprintf(_T("Reader supports card capture\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_BWT, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current block waiting time: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_CLK, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current clock rate (kHz): "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_CWT, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current character waiting time: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_D, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Bit rate conversion factor: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_EBC_ENCODING, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  auto nAttribSize{attrib.size()}; //NOLINT(clang-analyzer-deadcode.DeadStores)
  sValue = AttributeBinaryToString(attrib);
  if (sValue == _T("0"))
    _tprintf(_T("LRC error block control coding\n"));
  else if (sValue == _T("1"))
    _tprintf(_T("CRC error block control coding\n"));
  else
  {
    _tprintf(_T("EBC Encoding: "));
    PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
    _tprintf(_T("\n"));
  }

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_F, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Clock conversion factor: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_IFSC, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current byte size for information field size card: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_IFSD, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current byte size for information field size device: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_N, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current guard time: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_PROTOCOL_TYPE, dwAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current protocol type: %u\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

  nRet = card.GetAttrib(SCARD_ATTR_CURRENT_W, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Current work waiting time: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_DEFAULT_CLK, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Default clock rate (kHz): "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_DEFAULT_DATA_RATE, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Default data rate (bps): "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_DEVICE_FRIENDLY_NAME, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Reader display name: "));
  PrintByteArray(attrib, AttributeType::STRING_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_DEVICE_SYSTEM_NAME, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Reader system name: "));
  PrintByteArray(attrib, AttributeType::STRING_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_DEVICE_UNIT, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Vendor Device Unit: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  BYTE byAttrib{0};
  nRet = card.GetAttrib(SCARD_ATTR_ICC_INTERFACE_STATUS, byAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  if (byAttrib)
    _tprintf(_T("Electrical contact is active\n"));
  else
    _tprintf(_T("Electrical contact is not active\n"));

  nRet = card.GetAttrib(SCARD_ATTR_ICC_PRESENCE, byAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  if (byAttrib == 0)
    _tprintf(_T("Card not present\n"));
  else if (byAttrib == 1)
    _tprintf(_T("Card present but not swallowed\n"));
  else if (byAttrib == 2)
    _tprintf(_T("Card present\n"));
  else if (byAttrib == 4)
    _tprintf(_T("Card confiscated\n"));

  nRet = card.GetAttrib(SCARD_ATTR_ICC_TYPE_PER_ATR, byAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  if (byAttrib == 0)
    _tprintf(_T("Unknown smart card type\n"));
  else if (byAttrib == 1)
    _tprintf(_T("7816 Asynchronous smart card\n"));
  else if (byAttrib == 2)
    _tprintf(_T("7816 Synchronous smart card\n"));

  nRet = card.GetAttrib(SCARD_ATTR_MAX_CLK, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Maximum clock rate (kHz): "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_MAX_DATA_RATE, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Maximum data rate (bps): "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_MAX_IFSD, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Maximum bytes for information file size device: "));
  PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_POWER_MGMT_SUPPORT, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  nAttribSize = attrib.size();
#pragma warning(suppress: 26446)
  if ((nAttribSize == 1) && (attrib[0] == 0))
    _tprintf(_T("Device does not support power down while card is present\n"));
  else if (nAttribSize == 1)
    _tprintf(_T("Device supports power down while card is present\n"));
  else
  {
    _tprintf(_T("Power management support: "));
    PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
    _tprintf(_T("\n"));
  }

  nRet = card.GetAttrib(SCARD_ATTR_PROTOCOL_TYPES, dwAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Supported protocol types: 0x%08X\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

  nRet = card.GetAttrib(SCARD_ATTR_VENDOR_IFD_SERIAL_NO, attrib);
  if (nRet == SCARD_S_SUCCESS)
  {
    _tprintf(_T("Vendor supplied interface serial number: "));
    PrintByteArray(attrib, AttributeType::BLOB_ATTRIBUTE);
    _tprintf(_T("\n"));
  }
  else
#pragma warning(suppress: 26472)
    _tprintf(_T("Could not obtain Vendor supplied interface serial number, Error:0x%08x\n"), static_cast<unsigned int>(nRet));

  nRet = card.GetAttrib(SCARD_ATTR_VENDOR_IFD_TYPE, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Vendor supplied interface serial type: "));
  PrintByteArray(attrib, AttributeType::BLOB_ATTRIBUTE);
  _tprintf(_T("\n"));

  nRet = card.GetAttrib(SCARD_ATTR_VENDOR_IFD_VERSION, dwAttrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Vendor supplied device version: %u\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

  nRet = card.GetAttrib(SCARD_ATTR_VENDOR_NAME, attrib); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(nRet == SCARD_S_SUCCESS);
  _tprintf(_T("Vendor name: "));
  PrintByteArray(attrib, AttributeType::ASCII_STRING_ATTRIBUTE);
  _tprintf(_T("\n"));

  DWORD dwBytesReturned{0};
  nRet = card.Control(1234, "1234", 4, nullptr, 0, &dwBytesReturned); //NOLINT(clang-analyzer-deadcode.DeadStores)
  //nRet = card.SetAttrib(SCARD_ATTR_SUPRESS_T1_IFS_REQUEST, nullptr, 0);

  //Test out the SCard::CAPDURequest::Parse method
  SCard::CAPDURequest request5;
#pragma warning(suppress: 26490)
  bool bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03"), 3); //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-but-set-variable)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04"), 4); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05"), 5); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06"), 6); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06\x07"), 7); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06\x07\x08"), 8); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06\x07\x08\x09"), 9); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A"), 10); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26490)
  bParse = request5.Parse(reinterpret_cast<const BYTE*>("\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B"), 11); //NOLINT(clang-analyzer-deadcode.DeadStores)

  //Test out the SCard::CCard::Transmit method
  SCard::CAPDURequest request1{0x00, 0xB2, 0x01, 0x0C};
  SCard::CAPDURequest request2;
  request2 = request1;
  SCard::CAPDURequest request3{request1};
#pragma warning(suppress: 26490)
  SCard::CAPDURequest request4{0x00, 0xA4, 0x04, 0x00, reinterpret_cast<const BYTE*>("1PAY.SYS.DDF01"), 14}; //Lets try to select the Payment System Environment (PSE) application
  SCard::CAPDUResponse response1;
  SCard::CAPDUResponse response2{response1};
  response1 = std::move(response2);
  SCard::CAPDUResponse response3;
  response3 = response1;
  SCard::CAPDUResponse response5{std::move(response1)};
  SCard::CAPDUResponse response4;
  nRet = card.Transmit(request4, response4);
  if (nRet == SCARD_S_SUCCESS)
  {
#pragma warning(suppress: 26472)
    _tprintf(_T("Inserted card has responded to PSE select file command with %02X%02X response\n"), static_cast<int>(response4.m_bySW1), static_cast<int>(response4.m_bySW2));
    BYTE byLE{0};
    if (response4.IsSuccessfulWithNoFurtherQualification() || response4.IsSuccessfulWithLEIndicated(byLE))
      _tprintf(_T("Inserted card has a PSE application on it (i.e. Supports EMV Contact)\n"));
    else
      _tprintf(_T("Inserted card does not have a PSE application on it (i.e. Does not support EMV Contact)\n"));
  }
  else
#pragma warning(suppress: 26472)
    _tprintf(_T("Card failed to respond to Transmit request, Error:0x%08x\n"), static_cast<unsigned int>(nRet));

#pragma warning(suppress: 26490)
  request4 = SCard::CAPDURequest{0x00, 0xA4, 0x04, 0x00, reinterpret_cast<const BYTE*>("2PAY.SYS.DDF01"), 14}; //Lets try to select the Paypass Payment System Environment (PSE) application
  nRet = card.Transmit(request4, response4);
  if (nRet == SCARD_S_SUCCESS)
  {
#pragma warning(suppress: 26472)
    _tprintf(_T("Inserted card has responded to PPSE select file command with %02X%02X response\n"), static_cast<int>(response4.m_bySW1), static_cast<int>(response4.m_bySW2));
    BYTE byLE{0};
    if (response4.IsSuccessfulWithNoFurtherQualification() || response4.IsSuccessfulWithLEIndicated(byLE))
      _tprintf(_T("Inserted card has a PPSE application on it (i.e Supports EMV Contactless)\n"));
    else
      _tprintf(_T("Inserted card does not have a PPSE application on it (i.e. Does not support EMV Contactless)\n"));
  }
  else
#pragma warning(suppress: 26472)
    _tprintf(_T("Card failed to respond to Transmit request, Error:0x%08x\n"), static_cast<unsigned int>(nRet));

  card.Disconnect(card.GetDisconnectDisposition());
  ctx4.Close();

  try
  {
    //Create the context
    SCard::CContext2 ctx2;
    ctx2.Create();
    h = ctx2.Detach();
    ctx2.Attach(h);
    SCard::CContext2 ctx8{std::move(ctx2)};
    ctx2.Attach(ctx8.Detach());
#pragma warning(suppress: 26477)
    ATLASSERT(ctx2.IsValid() == SCARD_S_SUCCESS);
    SCard::CContext2 ctx9{std::move(ctx2)};
    SCard::CContext2 ctx10{ctx8.Detach()};

    SCard::CContext2 ctx7;
    ctx7 = std::move(ctx9);
    ctx7.ListReaders(SCARD_ALL_READERS, readers);
    i = 0;
    for (const auto& reader : readers)
    {
      _tprintf(_T("Found Reader: %s\n"), reader.c_str());

      if (i == 0)
        sFirstReader = reader;
      try
      {
        ctx7.GetReaderDeviceInstanceId(reader.c_str(), sValue);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.ListReadersWithDeviceInstanceId(sValue.c_str(), readers2);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      ++i;
    }

    ctx7.ListReaderGroups(readerGroups);
    for (const auto& readerGroup : readerGroups)
      _tprintf(_T("Found Reader group: %s\n"), readerGroup.c_str());

    //uncomment and modify the following line if you would like to connect to a specific smart card reader
    //sFirstReader = _T("SCM Microsystems Inc. SDI011G Smart Card Reader 0");

    if (sFirstReader.length() == 0)
    {
      _tprintf(_T("No readers found on this computer, giving up!\n"));
      return 1;
    }

    memset(&srs, 0, sizeof(srs));
    srs.szReader = sFirstReader.c_str();
    srs.pvUserData = nullptr;
    srs.dwCurrentState = SCARD_STATE_UNAWARE;
    ctx7.LocateCards(_T("Identity Device (NIST SP 800-73 [PIV])\0"), &srs, 1);

    atrMask.cbAtr = 23;
#pragma warning(suppress: 26485)
    memcpy_s(atrMask.rgbAtr, sizeof(atrMask.rgbAtr), "\x3B\xEE\x00\x00\x81\x31\xFE\x35\x00\x31\xC0\x65\xBC\xD0\x02\x01\x06\x71\xD6\x8C\x61\x22\x00", 23);
#pragma warning(suppress: 26485)
    memset(atrMask.rgbMask, 0xFF, sizeof(atrMask.rgbMask));
    ctx7.LocateCardsByATR(&atrMask, 1, &srs, 1);

    ctx7.IntroduceCardType(_T("PJ's Card"), &MyGuid, nullptr, 0, MyATR.data(), MyATRMask.data(), sizeof(MyATR));

    ctx7.ForgetCardType(_T("PJ's Card"));

    ctx7.IntroduceReader(_T("PJs Reader"), _T("Broadcom Corp Contacted SmartCard 0"));

    ctx7.ForgetReader(_T("PJs Reader"));

    ctx7.ListCards(nullptr, nullptr, 0, cards);
    for (const auto& cardX : cards)
    {
      LPCTSTR pszCard{cardX.c_str()};
      _tprintf(_T("Found Card: %s\n"), pszCard);

      try
      {
        ctx7.GetProviderName(pszCard, SCARD_PROVIDER_PRIMARY, sValue);
        _tprintf(_T(" Primary provider: %s\n"), sValue.c_str());
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.GetProviderName(pszCard, SCARD_PROVIDER_CSP, sValue);
        _tprintf(_T(" CSP provider: %s\n"), sValue.c_str());
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.GetProviderName(pszCard, SCARD_PROVIDER_KSP, sValue);
        _tprintf(_T(" KSP provider: %s\n"), sValue.c_str());
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.GetProviderName(pszCard, SCARD_PROVIDER_CARD_MODULE, sValue);
        _tprintf(_T(" Card module: %s\n"), sValue.c_str());
      }
      catch(SCard::CException& /*e*/)
      {
      }

      GUID gProviderId{};
      try
      {
        ctx7.GetProviderId(pszCard, &gProviderId);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      SCard::GUIDArray interfaces;
      ctx7.ListInterfaces(pszCard, interfaces);

      ctx7.IntroduceReaderGroup(_T("PJs Group"));
      //ctx7.AddReaderToGroup(_T("Broadcom Corp Contacted SmartCard 0\0"), _T("PJs Group"));
      //ctx7.RemoveReaderFromGroup(_T("Broadcom Corp Contacted SmartCard 0\0"), _T("PJs Group"));

      ctx7.ListReaderGroups(readerGroups);
      for (const auto& readerGroup : readerGroups)
        _tprintf(_T("Found Reader group: %s\n"), readerGroup.c_str());

      ctx7.ForgetReaderGroup(_T("PJs Group"));
      ctx7.Cancel();
      //ctx7.ForgetCardType(_T("DOES NOT EXIST"));

      try
      {
        ctx7.Audit(SCARD_AUDIT_CHV_FAILURE);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      dwDeviceTypeId = 0;
      try
      {
        ctx7.GetDeviceTypeId(sFirstReader.c_str(), &dwDeviceTypeId);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.GetReaderDeviceInstanceId(sFirstReader.c_str(), sDeviceInstanceId);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.ListReadersWithDeviceInstanceId(sDeviceInstanceId.c_str(), readers);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        ctx7.GetReaderIcon(sFirstReader.c_str(), icon);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      try
      {
        status = UuidCreate(&uuid); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(status == RPC_S_OK);
#pragma warning(suppress: 26465 26490 26492)
        ctx7.WriteCache(&uuid, 1, const_cast<LPTSTR>(_T("PJ_Value1")), const_cast<PBYTE>(reinterpret_cast<LPCBYTE>("1234")), 4);
        std::array<BYTE, 4> buffer{};
#pragma warning(suppress: 26472)
        auto dwSize{static_cast<DWORD>(buffer.size())};
#pragma warning(suppress: 26465 26492)
        ctx7.ReadCache(&uuid, 1, const_cast<LPTSTR>(_T("PJ_Value1")), buffer.data(), &dwSize);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      //ctx7.Close();

      _tprintf(_T("Checking state of the first reader for the next 10 seconds\n"));
      for (int j=0; j<40; j++)
      {
        srs.dwCurrentState = SCARD_STATE_UNAWARE;
        ctx7.GetStatusChange(INFINITE, &srs, 1);

        if (srs.dwEventState & SCARD_STATE_IGNORE)
          _tprintf(_T("Reader ignored,"));
        if (srs.dwEventState & SCARD_STATE_CHANGED)
          _tprintf(_T("Reader state has changed,"));
        if (srs.dwEventState & SCARD_STATE_UNKNOWN)
          _tprintf(_T("Reader name is not recognized,"));
        if (srs.dwEventState & SCARD_STATE_UNAVAILABLE)
          _tprintf(_T("Reader state not available,"));
        if (srs.dwEventState & SCARD_STATE_EMPTY)
          _tprintf(_T("No card in the reader,"));
        if (srs.dwEventState & SCARD_STATE_PRESENT)
          _tprintf(_T("Card in the reader,"));
        if (srs.dwEventState & SCARD_STATE_ATRMATCH)
          _tprintf(_T("Matching ATR found,"));
        if (srs.dwEventState & SCARD_STATE_EXCLUSIVE)
          _tprintf(_T("Card in the reader is allocated for exclusive use by another application,"));
        if (srs.dwEventState & SCARD_STATE_INUSE)
          _tprintf(_T("Card in use in the reader but may be connected to in shared mode,"));
        if (srs.dwEventState & SCARD_STATE_MUTE)
          _tprintf(_T("There is an unresponsive card in the reader,"));

        if (srs.cbAtr)
        {
          _tprintf(_T("ATR of card: "));
          for (DWORD k=0; k<srs.cbAtr; k++)
#pragma warning(suppress: 26446 26472 26482)
            _tprintf(_T("%02X"), static_cast<int>(srs.rgbAtr[k]));
          _tprintf(_T("\n"));
        }
        else
          _tprintf(_T("\n"));

        //The current state is now the event state
        srs.dwCurrentState = srs.dwEventState;

        Sleep(250);
      }

      SCard::CCard2 card5;
      try
      {
        _tprintf(_T("Attempting to connect to reader %s\n"), sFirstReader.c_str());
        card5.Connect(ctx7, sFirstReader.c_str());
      }
#pragma warning(suppress: 26496)
      catch(SCard::CException& e)
      {
#pragma warning(suppress: 26472)
        _tprintf(_T("Could not connect to the card in the reader, giving up!, Error:0x%08X\n"), static_cast<unsigned int>(e.m_nValue));
        return 1;
      }

      SCard::CCard2 card3;
      hCard = card3.Detach();
      SCard::CCard2 card4;
      card4.Attach(hCard);
      card4 = std::move(card3);
      card3 = std::move(card4);
      SCard::CCard2 card6{std::move(card4)};
      SCard::CCard2 card7{card6.Detach(), SCARD_LEAVE_CARD};

      dwState = 0;
      dwProtocol = 0;
      card5.Status(readerNames, &dwState, &dwProtocol, Atr);
      switch (dwState)
      {
        case SCARD_ABSENT:
        {
          _tprintf(_T("Card absent\n"));
          break;
        }
        case SCARD_PRESENT:
        {
          _tprintf(_T("Card present\n"));
          break;
        }
        case SCARD_SWALLOWED:
        {
          _tprintf(_T("Card swallowed\n"));
          break;
        }
        case SCARD_POWERED:
        {
          _tprintf(_T("Power has power\n"));
          break;
        }
        case SCARD_NEGOTIABLE:
        {
          _tprintf(_T("Card reset and waiting PTS negotiation\n"));
          break;
        }
        case SCARD_SPECIFIC:
        {
          _tprintf(_T("Card has specific communication protocols set\n"));
          break;
        }
        default:
        {
          _tprintf(_T("Unknown or unexpected card state\n"));
          break;
        }
      }
      PrintATR(Atr);

      std::array<TCHAR, 1024> mszReaderArray{};
#pragma warning(suppress: 26472)
      auto dwReaderNameSize{static_cast<DWORD>(mszReaderArray.size())};
      std::array<BYTE, 32> ATR2{};
#pragma warning(suppress: 26472)
      auto dwATRSize{static_cast<DWORD>(ATR2.size())};
      card5.Status(mszReaderArray.data(), &dwReaderNameSize, &dwState, &dwProtocol, ATR2.data(), &dwATRSize);

      dwActiveProtocol = card5.GetActiveProtocol(); //NOLINT(clang-analyzer-deadcode.DeadStores)

      pbAttr = nullptr;
      dwByte = SCARD_AUTOALLOCATE;
#pragma warning(suppress: 26490)
      card5.GetAttrib(SCARD_ATTR_DEVICE_SYSTEM_NAME, reinterpret_cast<LPBYTE>(&pbAttr), &dwByte);
      ctx7.FreeMemory(pbAttr);

      card5.BeginTransaction();

      card5.EndTransaction();

      _tprintf(_T("Transmit Count: %u\n"), card5.GetTransmitCount()); //NOLINT(clang-diagnostic-format)

      card5.Reconnect();

      card5.GetAttrib(SCARD_ATTR_ATR_STRING, attrib);
      _tprintf(_T("ATR of card: "));
      for (const auto& ATRChar : attrib)
#pragma warning(suppress: 26472)
        _tprintf(_T("%02X"), static_cast<int>(ATRChar));
      _tprintf(_T("\n"));

      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_CHANNEL_ID);
      if (dwAttrib & 0x80000)
        _tprintf(_T("Reader is SCSI based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else if (dwAttrib & 0x40000)
        _tprintf(_T("Reader is PS/2 Keyboard based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else if (dwAttrib & 0x20000)
        _tprintf(_T("Reader is Parallel port based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else if (dwAttrib & 0x10000)
        _tprintf(_T("Reader is Serial port based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else if (dwAttrib & 0x10000)
        _tprintf(_T("Reader is IDE based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else if (dwAttrib & 0x200000)
        _tprintf(_T("Reader is USB based, Channel number: %u\n"), dwAttrib & 0xFFFF); //NOLINT(clang-diagnostic-format)
      else
        _tprintf(_T("Reader Characteristics, Attribute value: 0x%08X\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_CHARACTERISTICS);
      if (dwAttrib == 0)
        _tprintf(_T("Reader has no special characteristics\n"));
      else if (dwAttrib & 0x1)
        _tprintf(_T("Reader supports card swallowing\n"));
      else if (dwAttrib & 0x2)
        _tprintf(_T("Reader supports card ejection\n"));
      else if (dwAttrib & 0x4)
        _tprintf(_T("Reader supports card capture\n"));
      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_CHARACTERISTICS); //NOLINT(clang-analyzer-deadcode.DeadStores)

      card5.GetAttrib(SCARD_ATTR_CURRENT_BWT, attrib);
      _tprintf(_T("Current block waiting time: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_CLK, attrib);
      _tprintf(_T("Current clock rate (kHz): "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_CWT, attrib);
      _tprintf(_T("Current character waiting time: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_D, attrib);
      _tprintf(_T("Bit rate conversion factor: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_EBC_ENCODING, attrib);
      nAttribSize = attrib.size(); //NOLINT(clang-analyzer-deadcode.DeadStores)
      sValue = AttributeBinaryToString(attrib);
      if (sValue == _T("0"))
        _tprintf(_T("LRC error block control coding\n"));
      else if (sValue == _T("1"))
        _tprintf(_T("CRC error block control coding\n"));
      else
      {
        _tprintf(_T("EBC Encoding: "));
        PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
        _tprintf(_T("\n"));
      }

      card5.GetAttrib(SCARD_ATTR_CURRENT_F, attrib);
      _tprintf(_T("Clock conversion factor: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_IFSC, attrib);
      _tprintf(_T("Current byte size for information field size card: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_IFSD, attrib);
      _tprintf(_T("Current byte size for information field size device: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_CURRENT_N, attrib);
      _tprintf(_T("Current guard time: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_CURRENT_PROTOCOL_TYPE);
      _tprintf(_T("Current protocol type: %u\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

      card5.GetAttrib(SCARD_ATTR_CURRENT_W, attrib);
      _tprintf(_T("Current work waiting time: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_DEFAULT_CLK, attrib);
      _tprintf(_T("Default clock rate (kHz): "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_DEFAULT_DATA_RATE, attrib);
      _tprintf(_T("Default data rate (bps): "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_DEVICE_FRIENDLY_NAME, attrib);
      _tprintf(_T("Reader display name: "));
      PrintByteArray(attrib, AttributeType::STRING_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_DEVICE_SYSTEM_NAME, attrib);
      _tprintf(_T("Reader system name: "));
      PrintByteArray(attrib, AttributeType::STRING_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_DEVICE_UNIT, attrib);
      _tprintf(_T("Vendor Device Unit: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      byAttrib = card5.GetAttribBYTE(SCARD_ATTR_ICC_INTERFACE_STATUS);
      if (byAttrib)
        _tprintf(_T("Electrical contact is active\n"));
      else
        _tprintf(_T("Electrical contact is not active\n"));

      byAttrib = card5.GetAttribBYTE(SCARD_ATTR_ICC_PRESENCE);
      if (byAttrib == 0)
        _tprintf(_T("Card not present\n"));
      else if (byAttrib == 1)
        _tprintf(_T("Card present but not swallowed\n"));
      else if (byAttrib == 2)
        _tprintf(_T("Card present\n"));
      else if (byAttrib == 4)
        _tprintf(_T("Card confiscated\n"));

      byAttrib = card5.GetAttribBYTE(SCARD_ATTR_ICC_TYPE_PER_ATR);
      if (byAttrib == 0)
        _tprintf(_T("Unknown smart card type\n"));
      else if (byAttrib == 1)
        _tprintf(_T("7816 Asynchronous smart card\n"));
      else if (byAttrib == 2)
        _tprintf(_T("7816 Synchronous smart card\n"));

      card5.GetAttrib(SCARD_ATTR_MAX_CLK, attrib);
      _tprintf(_T("Maximum clock rate (kHz): "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_MAX_DATA_RATE, attrib);
      _tprintf(_T("Maximum data rate (bps): "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_MAX_IFSD, attrib);
      _tprintf(_T("Maximum bytes for information file size device: "));
      PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
      _tprintf(_T("\n"));

      card5.GetAttrib(SCARD_ATTR_POWER_MGMT_SUPPORT, attrib);
      nAttribSize = attrib.size();
#pragma warning(suppress: 26446)
      if ((nAttribSize == 1) && (attrib[0] == 0))
        _tprintf(_T("Device does not support power down while card is present\n"));
      else if (nAttribSize == 1)
        _tprintf(_T("Device supports power down while card is present\n"));
      else
      {
        _tprintf(_T("Power management support: "));
        PrintByteArray(attrib, AttributeType::BLOB_OR_INTEGER_ATTRIBUTE);
        _tprintf(_T("\n"));
      }

      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_PROTOCOL_TYPES);
      _tprintf(_T("Supported protocol types: 0x%08X\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

      try
      {
        card5.GetAttrib(SCARD_ATTR_VENDOR_IFD_SERIAL_NO, attrib);
        _tprintf(_T("Vendor supplied interface serial number: "));
        PrintByteArray(attrib, AttributeType::BLOB_ATTRIBUTE);
        _tprintf(_T("\n"));
      }
#pragma warning(suppress: 26496)
      catch(SCard::CException& e)
      {
#pragma warning(suppress: 26472)
        _tprintf(_T("Could not obtain Vendor supplied interface serial number, Error:0x%08x\n"), static_cast<unsigned int>(e.m_nValue));
      }

      card5.GetAttrib(SCARD_ATTR_VENDOR_IFD_TYPE, attrib);
      _tprintf(_T("Vendor supplied interface serial type: "));
      PrintByteArray(attrib, AttributeType::BLOB_ATTRIBUTE);
      _tprintf(_T("\n"));

      dwAttrib = card5.GetAttribDWORD(SCARD_ATTR_VENDOR_IFD_VERSION);
      _tprintf(_T("Vendor supplied device version: %u\n"), dwAttrib); //NOLINT(clang-diagnostic-format)

      card5.GetAttrib(SCARD_ATTR_VENDOR_NAME, attrib);
      _tprintf(_T("Vendor name: "));
      PrintByteArray(attrib, AttributeType::ASCII_STRING_ATTRIBUTE);
      _tprintf(_T("\n"));

      try
      {
        card5.Control(1234, "1234", 4, nullptr, 0, &dwBytesReturned);
        card5.SetAttrib(SCARD_ATTR_SUPRESS_T1_IFS_REQUEST, nullptr, 0);
      }
      catch(SCard::CException& /*e*/)
      {
      }

      request1 = SCard::CAPDURequest{0x00, 0xB2, 0x01, 0x0C};
      request2 = request1;
      request3 = request1;
#pragma warning(suppress: 26490)
      request4 = SCard::CAPDURequest{0x00, 0xA4, 0x04, 0x00, reinterpret_cast<const BYTE*>("1PAY.SYS.DDF01"), 14}; //Lets try to select the Payment System Environment (PSE) application
      response2 = response1;
      response3 = response1;
      try
      {
        card5.Transmit(request4, response4);
#pragma warning(suppress: 26472)
        _tprintf(_T("Inserted card has responded to PSE select file command with %02X%02X response\n"), static_cast<int>(response4.m_bySW1), static_cast<int>(response4.m_bySW2));
        BYTE byLE{0};
        if (response4.IsSuccessfulWithNoFurtherQualification() || response4.IsSuccessfulWithLEIndicated(byLE))
          _tprintf(_T("Inserted card has a PSE application on it (i.e. Supports EMV Contact)\n"));
        else
          _tprintf(_T("Inserted card does not have a PSE application on it (i.e. Does not support EMV Contact)\n"));
      }
#pragma warning(suppress: 26496)
      catch(SCard::CException& e)
      {
#pragma warning(suppress: 26472)
        _tprintf(_T("Card failed to respond to Transmit request, Error:0x%08x\n"), static_cast<unsigned int>(e.m_nValue));
      }

#pragma warning(suppress: 26490)
      request4 = SCard::CAPDURequest{0x00, 0xA4, 0x04, 0x00, reinterpret_cast<const BYTE*>("2PAY.SYS.DDF01"), 14}; //Lets try to select the Paypass Payment System Environment (PSE) application
      try
      {
        card5.Transmit(request4, response4);
#pragma warning(suppress: 26472)
        _tprintf(_T("Inserted card has responded to PPSE select file command with %02X%02X response\n"), static_cast<int>(response4.m_bySW1), static_cast<int>(response4.m_bySW2));
        BYTE byLE{0};
        if (response4.IsSuccessfulWithNoFurtherQualification() || response4.IsSuccessfulWithLEIndicated(byLE))
          _tprintf(_T("Inserted card has a PPSE application on it (i.e Supports EMV Contactless)\n"));
        else
          _tprintf(_T("Inserted card does not have a PPSE application on it (i.e. Does not support EMV Contactless)\n"));
      }
#pragma warning(suppress: 26496)
      catch(SCard::CException& e)
      {
#pragma warning(suppress: 26472)
        _tprintf(_T("Card failed to respond to Transmit request, Error:0x%08X\n"), static_cast<unsigned int>(e.m_nValue));
      }

      ///*
      try
      {
        std::array<BYTE, 2> receiveBuffer{};
#pragma warning(suppress: 26472)
        auto dwResponseSize{static_cast<DWORD>(receiveBuffer.size())};
#pragma warning(suppress: 26490)
        card5.Transmit(SCARD_PCI_T0, reinterpret_cast<LPCBYTE>("\x00\xa4\x04\x00\x0e\x31\x50\x41\x59\x2e\x53\x59\x53\x2e\x44\x44\x46\x30\x31"), 19, nullptr, receiveBuffer.data(), &dwResponseSize);
      }
#pragma warning(suppress: 26496)
      catch(SCard::CException& e)
      {
#pragma warning(suppress: 26472)
        _tprintf(_T("Card failed to respond to Transmit request, Error:0x%08x\n"), static_cast<unsigned int>(e.m_nValue));
      }
      //*/

      card5.Disconnect(card5.GetDisconnectDisposition());
      //ctx7.Close();
    }
  }
#pragma warning(suppress: 26496)
  catch(SCard::CException& e)
  {
#pragma warning(suppress: 26472)
    _tprintf(_T("Unexpected SCard exception, Error Value:0x%08X\n"), static_cast<unsigned int>(e.m_nValue));
  }

  return 0;
}
